# Copyright (c) 2021-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers tools for `Optuna <https://optuna.readthedocs.io/en/stable/>`_.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[optuna]``
"""


import contextlib
import logging

import numpy as np
import optuna
from optuna.pruners import BasePruner
from optuna.study import StudyDirection
from scipy import stats

_logger = logging.getLogger(__name__)


class SignificanceRepeatedTrainingPruner(BasePruner):
    """Optuna pruner which uses statistical significance as an heuristic for decision-making.

    This is an Optuna :mod:`Pruner <optuna.pruners>` which uses statistical significance as
    an heuristic for decision-making. It prunes repeated trainings like in a cross validation.
    As the test method a `t-test <https://en.wikipedia.org/wiki/Student's_t-test>`_ is used.
    Our experiments have shown that an ``aplha`` value between 0.3 and 0.4 is reasonable.

    :mod:`Optuna's standard pruners <optuna.pruners>` assume that you only adjust the model once
    per hyperparameter set. Those pruners work on the basis of intermediate results. For example,
    once per epoch. In contrast, this pruner does not work on intermediate results but on the
    results of a cross validation or more precisely the results of the individual folds.

    Below is a minimalist example:

    .. testcode::

        from mltb2.optuna import SignificanceRepeatedTrainingPruner
        import logging
        import numpy as np
        import optuna
        from sklearn.datasets import load_iris
        from sklearn.model_selection import StratifiedKFold
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.metrics import accuracy_score

        # configure the logger to see the debug output from the pruner
        logging.getLogger().addHandler(logging.StreamHandler())
        logging.getLogger("mltb2.optuna").setLevel(logging.DEBUG)

        dataset = load_iris()

        x, y = dataset['data'], dataset['target']

        def train(trial):
            parameter = {
                'min_samples_split': trial.suggest_int('min_samples_split', 2, 20),
                'n_estimators': trial.suggest_int('n_estimators', 20, 100),
            }

            validation_result_list = []

            skf = StratifiedKFold(n_splits=10)
            for fold_index, (train_index, val_index) in enumerate(skf.split(x, y)):
                X_train, X_val = x[train_index], x[val_index]
                y_train, y_val = y[train_index], y[val_index]

                rf = RandomForestClassifier(**parameter)
                rf.fit(X_train, y_train)
                y_pred = rf.predict(X_val)

                acc = accuracy_score(y_val, y_pred)
                validation_result_list.append(acc)

                # report result of this fold
                trial.report(acc, fold_index)

                # check if we should prune
                if trial.should_prune():
                    # prune here - we are done with this CV
                    break

            return np.mean(validation_result_list)

        study = optuna.create_study(
            # storage="sqlite:///optuna.db",  # we use in-memory storage here
            study_name="iris_cv",
            direction="maximize",
            load_if_exists=True,
            sampler=optuna.samplers.TPESampler(multivariate=True),
            # add pruner to optuna
            pruner=SignificanceRepeatedTrainingPruner(
                alpha=0.4,
                n_warmup_steps=4,
            )
        )

        study.optimize(train, n_trials=10)

    Args:
        alpha: The alpha level for the statistical significance test.
            The larger this value is, the more aggressively this pruner works.
            The smaller this value is, the stronger the statistical difference between the two
            distributions must be for Optuna to prune.
            ``alpha`` must be ``0 < alpha < 1``.
            Our experiments have shown that an ``aplha`` value between 0.3 and 0.4 is reasonable.
        n_warmup_steps: Pruning is disabled until the trial reaches or exceeds the given number
            of steps.
    """

    def __init__(self, alpha: float = 0.1, n_warmup_steps: int = 4) -> None:
        # input value check
        if n_warmup_steps < 0:
            raise ValueError(f"'n_warmup_steps' must not be negative! n_warmup_steps: {n_warmup_steps}")
        if alpha >= 1:
            raise ValueError(f"'alpha' must be smaller than 1! {alpha}")
        if alpha <= 0:
            raise ValueError(f"'alpha' must be greater than 0! {alpha}")

        self.n_warmup_steps = n_warmup_steps
        self.alpha = alpha

    def prune(self, study: optuna.study.Study, trial: optuna.trial.FrozenTrial) -> bool:  # noqa: D102
        # get best tial - best trial is not available for first trial
        best_trial = None
        with contextlib.suppress(ValueError):
            best_trial = study.best_trial

        if best_trial is not None:
            trial_intermediate_values = list(trial.intermediate_values.values())

            _logger.debug("trial_intermediate_values: %s", trial_intermediate_values)

            # wait until the trial reaches or exceeds n_warmup_steps number of steps
            if len(trial_intermediate_values) >= self.n_warmup_steps:
                trial_mean = np.mean(trial_intermediate_values)

                best_trial_intermediate_values = list(best_trial.intermediate_values.values())
                best_trial_mean = np.mean(best_trial_intermediate_values)

                _logger.debug("trial_mean: %s", trial_mean)
                _logger.debug("best_trial_intermediate_values: %s", best_trial_intermediate_values)
                _logger.debug("best_trial_mean: %s", best_trial_mean)

                if (trial_mean < best_trial_mean and study.direction == StudyDirection.MAXIMIZE) or (
                    trial_mean > best_trial_mean and study.direction == StudyDirection.MINIMIZE
                ):
                    pvalue = stats.ttest_ind(
                        trial_intermediate_values,
                        best_trial_intermediate_values,
                    ).pvalue

                    _logger.debug("pvalue: %s", pvalue)

                    if pvalue < self.alpha:
                        _logger.info("We prune this trial. pvalue: %s", pvalue)

                        return True

                else:
                    _logger.debug("This trial is better than best trial - we do not check for pruning.")

            else:
                _logger.debug("This trial did not reach n_warmup_steps - we do no checks.")

        return False



# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""ArangoDB utils module.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[arangodb]``
"""


import gzip
from argparse import ArgumentParser
from contextlib import closing
from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence, Union

import jsonlines
from arango import ArangoClient
from arango.database import StandardDatabase
from dotenv import dotenv_values
from pandas import DataFrame
from tqdm import tqdm

from mltb2.db import AbstractBatchDataManager


def _check_config_keys(config: Dict[str, Optional[str]], expected_config_keys: Sequence[str]) -> None:
    """Check if all expected keys are in config.

    This is useful to check if a config file contains all necessary keys.
    """
    for expected_config_key in expected_config_keys:
        if expected_config_key not in config:
            raise ValueError(f"Config file must contain '{expected_config_key}'!")


@dataclass
class ArangoConnectionManager:
    """ArangoDB connection manager.

    Base class to manage / create ArangoDB connections.

    Args:
        hosts: ArangoDB host or hosts.
        db_name: ArangoDB database name.
        username: ArangoDB username.
        password: ArangoDB password.
    """

    hosts: Union[str, Sequence[str]]
    db_name: str
    username: str
    password: str

    def _arango_client_factory(self) -> ArangoClient:
        """Create an ArangoDB client."""
        arango_client = ArangoClient(hosts=self.hosts)
        return arango_client

    def _connection_factory(self, arango_client: ArangoClient) -> StandardDatabase:
        """Create an ArangoDB connection.

        Args:
            arango_client: ArangoDB client.
        """
        connection = arango_client.db(self.db_name, username=self.username, password=self.password)
        return connection


@dataclass
class ArangoBatchDataManager(AbstractBatchDataManager, ArangoConnectionManager):
    """ArangoDB implementation of the ``AbstractBatchDataManager``.

    Args:
        hosts: ArangoDB host or hosts.
        db_name: ArangoDB database name.
        username: ArangoDB username.
        password: ArangoDB password.
        collection_name: Documents from this collection are processed.
        attribute_name: This attribute is used to check if a document is already processed.
            If the attribute is not present in a document, the document is processed.
            If it is available the document is considered as already processed.
        batch_size: The batch size.
        aql_overwrite: AQL string to overwrite the default.
    """

    collection_name: str
    attribute_name: str
    batch_size: int = 20
    aql_overwrite: Optional[str] = None

    @classmethod
    def from_config_file(cls, config_file_name, aql_overwrite: Optional[str] = None):
        """Construct this from config file.

        The config file must contain these values:

        - ``hosts``
        - ``db_name``
        - ``username``
        - ``password``
        - ``collection_name``
        - ``attribute_name``
        - ``batch_size``

        Config file example:

        .. code-block::

            hosts="https://arangodb.com"
            db_name="my_ml_database"
            username="my_username"
            password="secret"
            collection_name="my_ml_data_collection"
            attribute_name="processing_metadata"
            batch_size=100

        Args:
            config_file_name: The config file name (path).
            aql_overwrite: AQL string to overwrite the default.
        """
        # load config file data
        arango_config = dotenv_values(config_file_name)

        # check if all necessary keys are in config file
        expected_config_file_keys = [
            "hosts",
            "db_name",
            "username",
            "password",
            "collection_name",
            "attribute_name",
            "batch_size",
        ]
        _check_config_keys(arango_config, expected_config_file_keys)

        return cls(
            hosts=arango_config["hosts"],  # type: ignore[arg-type]
            db_name=arango_config["db_name"],  # type: ignore[arg-type]
            username=arango_config["username"],  # type: ignore[arg-type]
            password=arango_config["password"],  # type: ignore[arg-type]
            collection_name=arango_config["collection_name"],  # type: ignore[arg-type]
            attribute_name=arango_config["attribute_name"],  # type: ignore[arg-type]
            batch_size=int(arango_config["batch_size"]),  # type: ignore[arg-type]
            aql_overwrite=aql_overwrite,
        )

    def load_batch(self) -> Sequence:
        """Load a batch of data from the ArangoDB database.

        Returns:
            The loaded batch of data.
        """
        with closing(self._arango_client_factory()) as arango_client:
            connection = self._connection_factory(arango_client)
            bind_vars = {
                "@coll": self.collection_name,
                "attribute": self.attribute_name,
                "batch_size": self.batch_size,
            }
            if self.aql_overwrite is None:
                aql = "FOR doc IN @@coll FILTER !HAS(doc, @attribute) LIMIT @batch_size RETURN doc"
            else:
                aql = self.aql_overwrite
            cursor = connection.aql.execute(
                aql,
                bind_vars=bind_vars,  # type: ignore[arg-type]
                batch_size=self.batch_size,
            )
            with closing(cursor) as closing_cursor:  # type: ignore[type-var]
                batch = closing_cursor.batch()  # type: ignore[union-attr]
        return batch  # type: ignore[return-value]

    def save_batch(self, batch: Sequence) -> None:
        """Save a batch of data to the ArangoDB database.

        Args:
            batch: The batch of data to save.
        """
        with closing(self._arango_client_factory()) as arango_client:
            connection = self._connection_factory(arango_client)
            collection = connection.collection(self.collection_name)
            collection.import_bulk(batch, on_duplicate="update")


def arango_collection_backup() -> None:
    """Commandline tool to do an ArangoDB backup of a collection.

    The backup is written to a gzip compressed JSONL file in the current working directory.
    Run ``arango-col-backup -h`` to get command line help.
    """
    # argument parsing
    description = (
        "ArangoDB backup of a collection. "
        "The backup is written to a gzip compressed JSONL file in the current working directory."
    )
    argument_parser = ArgumentParser(description=description)
    argument_parser.add_argument(
        "--conf", type=str, required=True, help="Config file containing 'hosts', 'db_name', 'username' and 'password'."
    )
    argument_parser.add_argument("--col", type=str, required=True, help="Collection name to backup.")
    args = argument_parser.parse_args()

    # load and check config file
    arango_config = dotenv_values(args.conf)
    expected_config_file_keys = ["hosts", "db_name", "username", "password"]
    _check_config_keys(arango_config, expected_config_file_keys)

    output_file_name = f"./{args.col}_backup.jsonl.gz"
    print(f"Writing backup to '{output_file_name}'...")

    with closing(ArangoClient(hosts=arango_config["hosts"])) as arango_client, gzip.open(  # type: ignore[arg-type]
        output_file_name, "w"
    ) as gzip_out:
        connection = arango_client.db(
            arango_config["db_name"],  # type: ignore[arg-type]
            arango_config["username"],  # type: ignore[arg-type]
            arango_config["password"],  # type: ignore[arg-type]
        )
        jsonlines_writer = jsonlines.Writer(gzip_out)  # type: ignore[arg-type]
        try:
            cursor = connection.aql.execute(
                "FOR doc IN @@coll RETURN doc",
                bind_vars={"@coll": args.col},
                batch_size=100,
                max_runtime=60 * 60,  # type: ignore[arg-type] # 1 hour
                stream=True,
            )
            for doc in tqdm(cursor):
                jsonlines_writer.write(doc)
        finally:
            cursor.close(ignore_missing=True)  # type: ignore[union-attr]


@dataclass
class ArangoImportDataManager(ArangoConnectionManager):
    """ArangoDB import tool to fill data into a collection.

    Args:
        hosts: ArangoDB host or hosts.
        db_name: ArangoDB database name.
        username: ArangoDB username.
        password: ArangoDB password.
    """

    @classmethod
    def from_config_file(cls, config_file_name):
        """Construct this from config file.

        The config file must contain at least these values:

        - ``hosts``
        - ``db_name``
        - ``username``
        - ``password``

        Config file example:

        .. code-block::

            hosts="https://arangodb.com"
            db_name="my_ml_database"
            username="my_username"
            password="secret"

        Args:
            config_file_name: The config file name (path).
        """
        # load config file data
        arango_config = dotenv_values(config_file_name)

        # check if all necessary keys are in config file
        expected_config_file_keys = [
            "hosts",
            "db_name",
            "username",
            "password",
        ]
        _check_config_keys(arango_config, expected_config_file_keys)

        return cls(
            hosts=arango_config["hosts"],
            db_name=arango_config["db_name"],
            username=arango_config["username"],
            password=arango_config["password"],
        )

    def import_dicts(
        self, dicts: Sequence[Dict[str, Any]], collection_name: str, create_collection: bool = False
    ) -> None:
        """Import data to ArangoDB.

        Args:
            dicts: The data to import.
            collection_name: The collection name to import to.
            create_collection: If ``True`` the collection is created if it does not exist.
        Raises:
            arango.exceptions.DocumentInsertError: If import fails.
        """
        with closing(self._arango_client_factory()) as arango_client:
            connection = self._connection_factory(arango_client)

            # get (or create) collection
            if not connection.has_collection(collection_name):
                if create_collection:
                    collection = connection.create_collection(collection_name)
                else:
                    raise ValueError(
                        f"Collection '{collection_name}' does not exist! "
                        "Create it or specify 'create_collection=True'."
                    )
            else:
                collection = connection.collection(collection_name)

            collection.import_bulk(  # type: ignore[union-attr]
                dicts,
                halt_on_error=True,
                details=False,
                overwrite=False,
                on_duplicate="error",
                sync=True,
                batch_size=100,
            )

    def import_dataframe(self, dataframe: DataFrame, collection_name: str, create_collection: bool = False) -> None:
        """Import Pandas data to ArangoDB.

        Args:
            dataframe: The Pandas data to import.
            collection_name: The collection name to import to.
            create_collection: If ``True`` the collection is created if it does not exist.
        Raises:
            arango.exceptions.DocumentInsertError: If import fails.
        """
        dicts = dataframe.to_dict(orient="records")
        self.import_dicts(dicts, collection_name, create_collection)



# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers tools for `fastText <https://fasttext.cc/docs/en/support.html>`_.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[fasttext]``
"""

import os
from dataclasses import dataclass, field
from typing import List, Optional

import fasttext
from fasttext.FastText import _FastText

from mltb2.files import fetch_remote_file, get_and_create_mltb2_data_dir


@dataclass
class FastTextLanguageIdentification:
    """Identify languages of a text."""

    model: _FastText = field(init=False, repr=False)

    def __post_init__(self):
        """Do post init."""
        self.model = fasttext.load_model(self.get_model_path_and_download())

    @staticmethod
    def get_model_path_and_download() -> str:
        """Get the model path and download it if needed.

        Returns:
            The full path to the downloaded model file.
        """
        model_filename = "lid.176.bin"
        mltb2_data_home = get_and_create_mltb2_data_dir()
        model_full_path = os.path.join(mltb2_data_home, model_filename)
        if not os.path.exists(model_full_path):
            model_url = "https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"
            sha256_checksum = "7e69ec5451bc261cc7844e49e4792a85d7f09c06789ec800fc4a44aec362764e"
            fetch_remote_file_path = fetch_remote_file(
                dirname=mltb2_data_home,
                filename=model_filename,
                url=model_url,
                sha256_checksum=sha256_checksum,
            )
            assert fetch_remote_file_path == model_full_path

        return model_full_path

    def __call__(self, text: str, num_lang: int = 10, always_detect_lang: Optional[List[str]] = None):
        """Identify languages of a given text.

        Args:
            text: the text for which the language is to be recognized
            num_lang: number of returned language probabilities
            always_detect_lang: A list of languages that should always be returned
                even if not detected. If the language is not detected, the probability
                is set to 0.0.
        Returns:
            A dict from language to probability.
            This dict contains no more than ``num_lang`` elements.
            So it is not guaranteed that the language you want to recognize is
            included in the dict. This is the case when the probability is very low.
            Possible languages are: ``af als am an ar arz as ast av az azb ba bar
            bcl be bg bh bn bo bpy br bs bxr ca cbk ce ceb ckb co cs cv cy da de
            diq dsb dty dv el eml en eo es et eu fa fi fr frr fy ga gd gl gn gom
            gu gv he hi hif hr hsb ht hu hy ia id ie ilo io is it ja jbo jv ka kk
            km kn ko krc ku kv kw ky la lb lez li lmo lo lrc lt lv mai mg mhr min
            mk ml mn mr mrj ms mt mwl my myv mzn nah nap nds ne new nl nn no oc or
            os pa pam pfl pl pms pnb ps pt qu rm ro ru rue sa sah sc scn sco sd sh
            si sk sl so sq sr su sv sw ta te tg th tk tl tr tt tyv ug uk ur uz vec
            vep vi vls vo wa war wuu xal xmf yi yo yue zh``
        """
        predictions = self.model.predict(text, k=num_lang)
        languages = predictions[0]
        probabilities = predictions[1]
        lang_to_prob = {lang[9:]: prob for lang, prob in zip(languages, probabilities)}
        if always_detect_lang is not None:
            for lang in always_detect_lang:
                if lang not in lang_to_prob:
                    lang_to_prob[lang] = 0.0
        return lang_to_prob



# Copyright (c) 2023 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Beautiful Soup and HTML specific tools.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[bs]``
"""

from typing import Any, Dict, Optional

import mdformat
from bs4 import BeautifulSoup
from markdownify import MarkdownConverter, markdownify


def extract_text(soup: BeautifulSoup, join_str: Optional[str] = None) -> str:
    """Extract the text from a BeautifulSoup object.

    Warning:
        This implementation has known issues with whitespace handling.

    Args:
        soup: The BeautifulSoup object to extract the text from.
        join_str: String to join the text parts with. Per default a space is used.
    Returns:
        Text from the BeautifulSoup object.
    """
    if join_str is None:
        join_str = " "
    texts = list(soup.stripped_strings)
    result: str = join_str.join(texts)
    return result


def extract_one(soup: BeautifulSoup, name=None, attrs: Optional[dict] = None, **kwargs: Dict[str, Any]) -> Any:
    """Extract exactly one specified element from a BeautifulSoup object.

    This function expacts that exactly only one result is found.
    Otherwise a RuntimeError is raised.

    Args:
        soup: The BeautifulSoup object to extract the element from.
        name: Name of the tag to extract.
        attrs: Attributes of the tag to extract.
        kwargs: Additional keyword arguments.
    Returns:
        The extracted BeautifulSoup element.
    Raises:
        RuntimeError: If not exactly one result is found.
    """
    if attrs is None:
        attrs = {}
    result = soup.find_all(name, attrs, **kwargs)
    if len(result) != 1:
        raise RuntimeError(f"Expected exactly one result, but got {len(result)}!")
    result = result[0]
    return result


def extract_all(soup: BeautifulSoup, name=None, attrs: Optional[dict] = None, **kwargs: Dict[str, Any]) -> Any:
    """Extract all specified elements from a BeautifulSoup object.

    Args:
        soup: The BeautifulSoup object to extract the elements from.
        name: Name of the tag to extract.
        attrs: Attributes of the tag to extract.
        kwargs: Additional keyword arguments.
    Returns:
        The extracted BeautifulSoup elements.
    """
    if attrs is None:
        attrs = {}
    result = soup.find_all(name, attrs, **kwargs)
    return result


def remove_all(soup: BeautifulSoup, name=None, attrs: Optional[dict] = None, **kwargs: Dict[str, Any]) -> None:
    """Remove all specified elements from a BeautifulSoup object.

    The removal is done in place. Nothing is returned.

    Args:
        soup: The BeautifulSoup object to remove the elements from.
        name: Name of the tag(-s) to remove.
        attrs: Attributes of the tag(-s) to remove.
        kwargs: Additional keyword arguments.
    """
    if attrs is None:
        attrs = {}
    result = soup.find_all(name, attrs, **kwargs)
    for r in result:
        r.decompose()


def soup_to_md(soup: BeautifulSoup, mdformat_options: Optional[dict] = None) -> str:
    """Convert a BeautifulSoup object to Markdown.

    The default mdformat options are:

    - ``number=True``: apply consecutive numbering to ordered lists
    - ``wrap="no"``: paragraph word wrap mode
    - ``end-of-line="lf"``: use LF as line ending

    See Also:
        The `mdformat Options
        <https://mdformat.readthedocs.io/en/stable/users/installation_and_usage.html#options>`_.
    Args:
        soup: BeautifulSoup object.
        mdformat_options: Options for mdformat.
    Returns:
        The Markdown text.
    """
    if mdformat_options is None:
        mdformat_options = {"number": True, "wrap": "no"}
    text = MarkdownConverter().convert_soup(soup)
    text = mdformat.text(text, options=mdformat_options)
    return text


def html_to_md(html: str, mdformat_options: Optional[dict] = None) -> str:
    """Convert HTML to Markdown.

    The default mdformat options are:

    - ``number=True``: apply consecutive numbering to ordered lists
    - ``wrap="no"``: paragraph word wrap mode
    - ``end-of-line="lf"``: use LF as line ending

    See Also:
        The `mdformat Options
        <https://mdformat.readthedocs.io/en/stable/users/installation_and_usage.html#options>`_.
    Args:
        html: HTML text.
        mdformat_options: Options for mdformat.
    Returns:
        The Markdown text.
    """
    if mdformat_options is None:
        mdformat_options = {"number": True, "wrap": "no"}
    text = markdownify(html)
    text = mdformat.text(text, options=mdformat_options)
    return text



# Copyright (c) 2023 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""File utils module.

This module provides utility functions for other modules.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[files]``
"""


import contextlib
import os
from typing import Optional

from platformdirs import user_data_dir
from sklearn.datasets._base import RemoteFileMetadata, _fetch_remote


def get_and_create_mltb2_data_dir(mltb2_base_data_dir: Optional[str] = None) -> str:
    """Return and create a data dir for mltb2.

    The exact directory is given by the ``mltb2_base_data_dir`` as the base folder
    and then the folder ``mltb2`` is appended.

    Args:
        mltb2_base_data_dir: The base data directory. If ``None`` the default
            user data directory is used. The default user data directory is
            determined by :func:`platformdirs.user_data_dir`.

    Returns:
        The directory path.
    """
    if mltb2_base_data_dir is None:
        mltb2_data_dir = user_data_dir(appname="mltb2")
    else:
        mltb2_data_dir = os.path.join(mltb2_base_data_dir, "mltb2")
    if not os.path.exists(mltb2_data_dir):
        os.makedirs(mltb2_data_dir)
    return mltb2_data_dir


def fetch_remote_file(dirname, filename, url: str, sha256_checksum: str) -> str:
    """Fetch a file from a remote URL.

    Args:
        dirname: the directory where the file will be saved
        filename: the filename under which the file will be saved
        url: the url of the file
        sha256_checksum: the sha256 checksum of the file
    Returns:
        Full path of the created file.
    Raises:
        IOError: if the sha256 checksum is wrong
    """
    remote = RemoteFileMetadata(filename=filename, url=url, checksum=sha256_checksum)
    try:
        fetch_remote_file_path = _fetch_remote(remote, dirname=dirname)
    except Exception:
        with contextlib.suppress(FileNotFoundError):
            os.remove(os.path.join(dirname, filename))
        raise
    return fetch_remote_file_path



# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Database utils module."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, Sequence


class AbstractBatchDataManager(ABC):
    """Abstract base class for batch processing of database data.

    This class (respectively an implementation of it) is intended to be
    used in conjunction with the :class:`BatchDataProcessor`.
    """

    @abstractmethod
    def load_batch(self) -> Sequence:
        """Load a batch of data from the database."""

    @abstractmethod
    def save_batch(self, batch: Sequence) -> None:
        """Save a batch of data to the database."""


@dataclass
class BatchDataProcessor:
    """Process batches of data from a database.

    Args:
        data_manager: The data manager to load and save batches of data.
        process_batch_callback: A callback function that processes one batch of data.
    """

    data_manager: AbstractBatchDataManager
    process_batch_callback: Callable[[Sequence], Sequence]

    def run(self) -> None:
        """Run the batch data processing.

        This is done until the data manager returns an empty batch.
        For each batch the ``process_batch_callback`` is called.
        Data is loaded by using an implementation of the :class:`AbstractBatchDataManager`.
        """
        while True:
            batch = self.data_manager.load_batch()
            if len(batch) == 0:
                break
            new_batch = self.process_batch_callback(batch)
            if len(new_batch) > 0:
                self.data_manager.save_batch(new_batch)



# Copyright (c) 2020-2024 Philip May
# Copyright (c) 2021 Sigrun May, Helmholtz-Zentrum für Infektionsforschung GmbH (HZI)
# Copyright (c) 2021 Sigrun May, Ostfalia Hochschule für angewandte Wissenschaften
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers tools for loading data.

The following tabular data sets from the biological and medical domain are supported:

- colon: `<http://genomics-pubs.princeton.edu/oncology/affydata/index.html>`_
- prostate: `<https://web.stanford.edu/~hastie/CASI_files/DATA/prostate.html>`_
- leukemia_big: `<https://web.stanford.edu/~hastie/CASI_files/DATA/leukemia.html>`_

After loading the data from the internet it is parsed, converted and
cached in the mltb2 data directory.
This data directory is determined by :func:`mltb2.files.get_and_create_mltb2_data_dir`.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[data]``
"""

import os
from hashlib import sha256
from io import StringIO
from typing import Optional, Tuple

import joblib
import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup

from mltb2.files import get_and_create_mltb2_data_dir


def _load_colon_data() -> pd.DataFrame:
    """Load colon data (not the labels).

    The data is loaded and parsed from the internet.
    Also see `<http://genomics-pubs.princeton.edu/oncology/affydata/index.html>`_.

    Returns:
        data as pandas DataFrame
    """
    # download data file
    url = "http://genomics-pubs.princeton.edu/oncology/affydata/I2000.html"
    page = requests.get(url, timeout=10)

    # check checksum of data file
    page_hash = sha256(page.content).hexdigest()
    assert page_hash == "74cc7b47d40a0fbca8dde05f42bcb799b7babad29ea634139a221bb4386b1c3d", page_hash

    soup = BeautifulSoup(page.content, "html.parser")
    page_text = soup.get_text()

    page_text_lines = page_text.splitlines()
    assert len(page_text_lines) >= 2000
    page_text_lines = [[float(s) for s in line.split()] for line in page_text_lines if len(line) > 20]
    assert len(page_text_lines) == 2000
    assert len(page_text_lines[0]) == 62

    data = np.array(page_text_lines).T
    data_df = pd.DataFrame(data)
    return data_df


def _load_colon_label() -> pd.Series:
    """Load colon label (not the data).

    The data is loaded and parsed from the internet.
    Also see `<http://genomics-pubs.princeton.edu/oncology/affydata/index.html>`_.

    Returns:
        labels as pandas Series
    """
    # download data file
    url = "http://genomics-pubs.princeton.edu/oncology/affydata/tissues.html"
    page = requests.get(url, timeout=10)

    # check checksum of data file
    page_hash = sha256(page.content).hexdigest()
    assert page_hash == "0c5b377c5dd5544d015bff479a4260d5ccf0bcf98657f600a1d37e34193e0f52", page_hash

    soup = BeautifulSoup(page.content, "html.parser")
    page_text = soup.get_text()
    page_text_lines = page_text.splitlines()

    label = []

    for line in page_text_lines:
        try:
            i = int(line)
            label.append(0 if i > 0 else 1)
        except ValueError:  # noqa: PERF203
            pass  # we ignore this

    assert len(label) == 62
    label_series = pd.Series(label)
    return label_series


def load_colon(mltb2_base_data_dir: Optional[str] = None) -> Tuple[pd.Series, pd.DataFrame]:
    """Load colon data.

    The data is loaded and parsed from the internet.
    Also see `<http://genomics-pubs.princeton.edu/oncology/affydata/index.html>`_.

    Args:
        mltb2_base_data_dir: The base data directory. If ``None`` the default
            user data directory is used. The default user data directory is
            determined by :func:`platformdirs.user_data_dir`.
    Returns:
        Tuple containing labels and data.
    """
    filename = "colon.pkl.gz"
    mltb2_data_home = get_and_create_mltb2_data_dir(mltb2_base_data_dir)
    full_path = os.path.join(mltb2_data_home, filename)
    if not os.path.exists(full_path):
        data_df = _load_colon_data()
        label_series = _load_colon_label()
        result = (label_series, data_df)
        joblib.dump(result, full_path, compress=("gzip", 3))
    else:
        result = joblib.load(full_path)

    return result


def load_prostate(mltb2_base_data_dir: Optional[str] = None) -> Tuple[pd.Series, pd.DataFrame]:
    """Load prostate data.

    The data is loaded and parsed from `<https://web.stanford.edu/~hastie/CASI_files/DATA/prostate.html>`_.

    Args:
        mltb2_base_data_dir: The base data directory. If ``None`` the default
            user data directory is used. The default user data directory is
            determined by :func:`platformdirs.user_data_dir`.
    Returns:
        Tuple containing labels and data.
    """
    filename = "prostate.pkl.gz"
    mltb2_data_home = get_and_create_mltb2_data_dir(mltb2_base_data_dir)
    full_path = os.path.join(mltb2_data_home, filename)
    if not os.path.exists(full_path):
        # download data file
        url = "https://web.stanford.edu/~hastie/CASI_files/DATA/prostmat.csv"
        page = requests.get(url, timeout=10)
        page_str = page.text

        # check checksum of data file
        page_hash = sha256(page_str.encode("utf-8")).hexdigest()
        assert page_hash == "f1ccfd3c9a837c002ec5d6489ab139c231739c3611189be14d15ca5541b92036", page_hash

        data_df = pd.read_csv(StringIO(page_str))
        data_df = data_df.T

        labels = []
        for label in data_df.index:
            if "control" in label:
                labels.append(0)
            elif "cancer" in label:
                labels.append(1)
            else:
                assert False, f"This must not happen! label: {label}"
        label_series = pd.Series(labels)
        assert len(label_series) == 102

        data_df = data_df.reset_index(drop=True)  # reset the index to default integer index
        assert data_df.shape == (102, 6033)

        result = (label_series, data_df)
        joblib.dump(result, full_path, compress=("gzip", 3))
    else:
        result = joblib.load(full_path)
    return result


def load_leukemia_big(mltb2_base_data_dir: Optional[str] = None) -> Tuple[pd.Series, pd.DataFrame]:
    """Load leukemia (big) data.

    The data is loaded and parsed from the internet.
    Also see `<https://web.stanford.edu/~hastie/CASI_files/DATA/leukemia.html>`_.

    Args:
        mltb2_base_data_dir: The base data directory. If ``None`` the default
            user data directory is used. The default user data directory is
            determined by :func:`platformdirs.user_data_dir`.
    Returns:
        Tuple containing labels and data.
    """
    filename = "leukemia_big.pkl.gz"
    mltb2_data_home = get_and_create_mltb2_data_dir(mltb2_base_data_dir)
    full_path = os.path.join(mltb2_data_home, filename)
    if not os.path.exists(full_path):
        # download data file
        url = "https://web.stanford.edu/~hastie/CASI_files/DATA/leukemia_big.csv"
        page = requests.get(url, timeout=10)
        page_str = page.text

        # check checksum of data file
        page_hash = sha256(page_str.encode("utf-8")).hexdigest()
        assert page_hash == "35e84928da625da0787efb31a451dedbdf390e821a94ef74b7b7ab6cab9466d4", page_hash

        data_df = pd.read_csv(StringIO(page_str))
        data_df = data_df.T

        labels = []
        for label in data_df.index:
            if "ALL" in label:
                labels.append(0)
            elif "AML" in label:
                labels.append(1)
            else:
                assert False, f"This must not happen! label: {label}"
        label_series = pd.Series(labels)
        assert len(label_series) == 72

        data_df = data_df.reset_index(drop=True)  # reset the index to default integer index
        assert data_df.shape == (72, 7128)

        result = (label_series, data_df)
        joblib.dump(result, full_path, compress=("gzip", 3))
    else:
        result = joblib.load(full_path)
    return result



# Copyright (c) 2023 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers `Hugging Face Transformers <https://huggingface.co/docs/transformers/index>`_ specific tools.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[transformers]``
"""

import os
from dataclasses import dataclass, field
from typing import Iterable, List, Union

import sklearn
import torch
from tqdm import tqdm
from transformers import AutoTokenizer
from transformers.tokenization_utils import PreTrainedTokenizerBase


@dataclass
class TransformersTokenCounter:
    """Count Transformers tokenizer tokens.

    Args:
        pretrained_model_name_or_path:
            The *model id* of a tokenizer hosted inside a model repo on huggingface.co or
            a path to a *directory* containing a tokenizer.
        show_progress_bar: Show a progressbar during processing.
    """

    pretrained_model_name_or_path: Union[str, os.PathLike]
    tokenizer: PreTrainedTokenizerBase = field(init=False, repr=False)
    show_progress_bar: bool = False

    def __post_init__(self):
        """Do post init."""
        self.tokenizer = AutoTokenizer.from_pretrained(self.pretrained_model_name_or_path)

    def __call__(self, text: Union[str, Iterable]) -> Union[int, List[int]]:
        """Count tokens for text.

        Args:
            text: The text for which the tokens are to be counted.
        Returns:
            The number of tokens if text was just a ``str``.
            If text is an ``Iterable`` then a ``list`` of number of tokens.
        """
        if isinstance(text, str):
            tokenized_text = self.tokenizer.tokenize(text)
            return len(tokenized_text)
        else:
            counts = []
            for t in tqdm(text, disable=not self.show_progress_bar):
                tokenized_text = self.tokenizer.tokenize(t)
                counts.append(len(tokenized_text))
            return counts


class LabeledDataset(torch.utils.data.Dataset):
    """Dataset with labes."""

    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):  # noqa: D105
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):  # noqa: D105
        return len(self.labels)


class KFoldLabeledDataset:
    """Utility to do k-fold cross-validation on ``LabeledDataset``."""

    def __init__(self, n_splits=7, n_repeats=1, random_state=None):
        self.n_splits = n_splits
        self.n_repeats = n_repeats
        self.random_state = random_state

    def split(self, labeled_dataset, stratification_labels=None):
        """Generates data splits of training and test set."""
        idxs = list(range(len(labeled_dataset)))
        if stratification_labels is None:  # no stratification wanted
            k_fold = sklearn.model_selection.RepeatedKFold(
                n_splits=self.n_splits,
                n_repeats=self.n_repeats,
                random_state=self.random_state,
            )
            k_fold_split = k_fold.split(idxs)
        else:  # stratification wanted
            k_fold = sklearn.model_selection.RepeatedStratifiedKFold(
                n_splits=self.n_splits,
                n_repeats=self.n_repeats,
                random_state=self.random_state,
            )
            k_fold_split = k_fold.split(idxs, stratification_labels)

        for train_idxs, test_idxs in k_fold_split:
            train = torch.utils.data.Subset(labeled_dataset, train_idxs)
            test = torch.utils.data.Subset(labeled_dataset, test_idxs)
            yield train, test



# Copyright (c) 2023 Philip May
# Copyright (c) 2023 Philip May, Deutsche Telekom AG
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers Hugging Face Transformers and SoMaJo specific tools.

This module is based on
`Hugging Face Transformers <https://huggingface.co/docs/transformers/index>`_ and
`SoMaJo <https://github.com/tsproisl/SoMaJo>`_.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[somajo_transformers]``
"""


from dataclasses import dataclass
from typing import List

from tqdm import tqdm

from mltb2.somajo import SoMaJoSentenceSplitter
from mltb2.transformers import TransformersTokenCounter


@dataclass
class TextSplitter:
    """Split the text into sections with a specified maximum token number.

    Does not divide words, but always whole sentences.

    Args:
        max_token: Maximum number of tokens per text section.
        somajo_sentence_splitter: The sentence splitter to be used.
        transformers_token_counter: The token counter to be used.
        show_progress_bar: Show a progressbar during processing.
        ignore_overly_long_sentences: If this is ``False`` an ``ValueError`` exception is
            raised if a sentence is longer than ``max_token``.
            If it is ``True``, then the sentence is simply ignored.
    """

    max_token: int
    somajo_sentence_splitter: SoMaJoSentenceSplitter
    transformers_token_counter: TransformersTokenCounter
    show_progress_bar: bool = False
    ignore_overly_long_sentences: bool = False

    def __call__(self, text: str) -> List[str]:
        """Split the text into sections.

        Args:
            text: The text to be split.
        Returns:
            The list of section splits.
        """
        sentences = self.somajo_sentence_splitter(text)
        counts = self.transformers_token_counter(sentences)

        assert len(sentences) == len(counts)  # type: ignore[arg-type]

        result_splits: List[str] = []
        current_sentences: List[str] = []
        current_count: int = 0
        for sentence, count in zip(tqdm(sentences, disable=not self.show_progress_bar), counts):  # type: ignore[arg-type]
            if count > self.max_token:
                if self.ignore_overly_long_sentences:
                    continue
                else:
                    raise ValueError("No sentence may be longer than 'max_token'.")
            if current_count + count <= self.max_token:
                current_count += count
                current_sentences.append(sentence)
            else:
                result_split = " ".join(current_sentences)
                result_splits.append(result_split)
                current_sentences = []
                current_count = 0
                current_count += count
                current_sentences.append(sentence)

        # add the rest
        if len(current_sentences) > 0:
            result_split = " ".join(current_sentences)
            result_splits.append(result_split)

        return result_splits



# Copyright (c) 2018-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Plot tools module.

This module is based on `Matplotlib <https://matplotlib.org/>`_.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[plot]``
"""

from typing import Optional

import matplotlib.pyplot as plt


# see https://matplotlib.org/api/_as_gen/matplotlib.axes.Axes.twinx.html
def twin_axes_timeseries_plot(
    values_1,
    label_1: str,
    values_2,
    label_2: str,
    start_timestep_number: int = 0,
    shift_1: int = 0,
    shift_2: int = 0,
    title: Optional[str] = None,
    label_x: str = "Step",
    color_1: str = "tab:red",
    color_2: str = "tab:blue",
):
    """Create twin axes timeseries plot.

    Plots two different timeseries curves in one diagram but two different y-axes.

    Hint:
        This function does not use `matplotlib.pyplot.plot`.

    Args:
        values_1: (``array_like``) Values for the first timeseries curve.
        label_1: Label for the first timeseries curve.
        values_2: (``array_like``) Values for the second timeseries curve.
        label_2: Label for the second timeseries curve.
        start_timestep_number: Number for first point in time. Default is 0.
        shift_1: Number of timesteps to shift the first timeseries curve.
            Can be positive or negative. Default is 0.
        shift_2: Number of timesteps to shift the second timeseries curve.
            Can be positive or negative. Default is 0.
        title: Title of the plot.
        label_x: Label for the x-axis (timeseries axis). Default is 'Step'.
        color_1: Color of first timeseries curve. Default is 'tab:red'.
        color_2: Color of second timeseries curve. Default is 'tab:blue'.
    """
    fig, ax1 = plt.subplots()

    if title is not None:
        plt.title(title)

    ax1.set_xlabel(label_x)

    ax1.set_ylabel(label_1, color=color_1)
    ax1.plot(
        range(start_timestep_number + shift_1, len(values_1) + start_timestep_number + shift_1),
        values_1,
        color=color_1,
    )
    ax1.tick_params(axis="y", labelcolor=color_1)

    ax2 = ax1.twinx()

    ax2.set_ylabel(label_2, color=color_2)
    ax2.plot(
        range(start_timestep_number + shift_2, len(values_2) + start_timestep_number + shift_2),
        values_2,
        color=color_2,
    )
    ax2.tick_params(axis="y", labelcolor=color_2)

    # otherwise the labels might be slightly clipped
    # see https://matplotlib.org/users/tight_layout_guide.html
    fig.tight_layout()


def boxplot(
    values,
    labels=None,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    vert: bool = True,
):
    """Prints one or more boxplots in a single diagram.

    Hint:
        This function does not use `matplotlib.pyplot.plot`.

    Args:
        values: Values for the boxplot(s).
        labels: Labels for the boxplot(s).
        title: Title of the plot.
        xlabel: Label for the x-axis.
        ylabel: Label for the y-axis.
        vert: If ``True`` (default), makes the boxes vertical.
            If ``False``, makes horizontal boxes.
    """
    _, ax = plt.subplots()

    if title is not None:
        ax.set_title(title)

    if xlabel is not None:
        ax.set(xlabel=xlabel)

    if ylabel is not None:
        ax.set(ylabel=ylabel)

    ax.boxplot(values, labels=labels, vert=vert)
    grid_axis = "y" if vert else "x"
    plt.grid(b=True, axis=grid_axis, linestyle="--")
    plt.xticks(rotation=90)


def boxplot_dict(
    values_dict,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    vert: bool = True,
):
    """Create boxplot form dictionary.

    Hint:
        This function does not use `matplotlib.pyplot.plot`.

    Args:
        values_dict: Dictionary with values for the boxplot(s).
        title: Title of the plot.
        xlabel: Label for the x-axis.
        ylabel: Label for the y-axis.
        vert: If ``True`` (default), makes the boxes vertical.
            If ``False``, makes horizontal boxes.
    """
    values = []
    labels = []

    for key, value in values_dict.items():
        values.append(value)
        labels.append(key)

    boxplot(values, labels=labels, title=title, xlabel=xlabel, ylabel=ylabel, vert=vert)


def save_last_figure(filename):
    """Saves the last plot made by Matplotlib.

    For jupyter notebooks this has to be called in the same cell that created the plot.
    """
    plt.savefig(filename, bbox_inches="tight")



# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""OpenAI specific module.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[openai]``
"""


from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, Union

import tiktoken
import yaml
from openai import AzureOpenAI, OpenAI
from openai.types.chat import ChatCompletion
from tiktoken.core import Encoding
from tqdm import tqdm


@dataclass
class OpenAiTokenCounter:
    """Count OpenAI tokens.

    Args:
        model_name:
            The OpenAI model name. Some examples:

            * ``gpt-4``
            * ``gpt-3.5-turbo``
            * ``text-davinci-003``
            * ``text-embedding-ada-002``

        show_progress_bar: Show a progressbar during processing.
    """

    model_name: str
    encoding: Encoding = field(init=False, repr=False)
    show_progress_bar: bool = False

    def __post_init__(self) -> None:
        """Do post init."""
        self.encoding = tiktoken.encoding_for_model(self.model_name)

    def __call__(self, text: Union[str, Iterable]) -> Union[int, List[int]]:
        """Count tokens for text.

        Args:
            text: The text for which the tokens are to be counted.
        Returns:
            The number of tokens if text was just a ``str``.
            If text is an ``Iterable`` then a ``list`` of number of tokens.
        """
        if isinstance(text, str):
            tokenized_text = self.encoding.encode(text)
            return len(tokenized_text)
        else:
            counts = []
            for t in tqdm(text, disable=not self.show_progress_bar):
                tokenized_text = self.encoding.encode(t)
                counts.append(len(tokenized_text))
            return counts


@dataclass
class OpenAiChatResult:
    """Result of an OpenAI chat completion.

    If you want to convert this to a ``dict`` use ``asdict(open_ai_chat_result)``
    from the ``dataclasses`` module.

    See Also:
        OpenAI API reference: `The chat completion object <https://platform.openai.com/docs/api-reference/chat/object>`_

    Args:
        content: the result of the OpenAI completion
        model: model name which has been used
        prompt_tokens: number of tokens of the prompt
        completion_tokens: number of tokens of the completion (``content``)
        total_tokens: number of total tokens (``prompt_tokens + content_tokens``)
        finish_reason: The reason why the completion stopped.

            * ``stop``: Means the API returned the full completion without running into any token limit.
            * ``length``: Means the API stopped the completion because of running into a token limit.
            * ``content_filter``: When content was omitted due to a flag from the OpenAI content filters.
            * ``tool_calls``: When the model called a tool.
            * ``function_call`` (deprecated): When the model called a function.

        completion_args: The arguments which have been used for the completion. Examples:

            * ``model``: always set
            * ``max_tokens``: only set if ``completion_kwargs`` contained ``max_tokens``
            * ``temperature``: only set if ``completion_kwargs`` contained ``temperature``
            * ``top_p``: only set if ``completion_kwargs`` contained ``top_p``
    """

    content: Optional[str] = None
    model: Optional[str] = None
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    total_tokens: Optional[int] = None
    finish_reason: Optional[str] = None
    completion_args: Optional[Dict[str, Any]] = None

    @classmethod
    def from_chat_completion(
        cls,
        chat_completion: ChatCompletion,
        completion_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """Construct this class from an OpenAI ``ChatCompletion`` object.

        Args:
            chat_completion: The OpenAI ``ChatCompletion`` object.
            completion_kwargs: The arguments which have been used for the completion.
        Returns:
            The constructed class.
        """
        result = {}
        result["completion_args"] = completion_kwargs
        chat_completion_dict = chat_completion.model_dump()
        result["model"] = chat_completion_dict.get("model")
        usage = chat_completion_dict.get("usage")
        if usage is not None:
            result["prompt_tokens"] = usage.get("prompt_tokens")
            result["completion_tokens"] = usage.get("completion_tokens")
            result["total_tokens"] = usage.get("total_tokens")
        choices = chat_completion_dict.get("choices")
        if choices is not None and len(choices) > 0:
            choice = choices[0]
            result["finish_reason"] = choice.get("finish_reason")
            message = choice.get("message")
            if message is not None:
                result["content"] = message.get("content")
        return cls(**result)  # type: ignore[arg-type]


@dataclass
class OpenAiChat:
    """Tool to interact with OpenAI chat models.

    This also be constructed with :meth:`~OpenAiChat.from_yaml`.

    See Also:
        OpenAI API reference: `Create chat completion <https://platform.openai.com/docs/api-reference/chat/create>`_

    Args:
        api_key: The OpenAI API key.
        model: The OpenAI model name.
    """

    api_key: str
    model: str
    client: Union[OpenAI, AzureOpenAI] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        """Do post init."""
        self.client = OpenAI(api_key=self.api_key)

    @classmethod
    def from_yaml(cls, yaml_file):
        """Construct this class from a yaml file.

        Args:
            yaml_file: The yaml file.
        Returns:
            The constructed class.
        """
        with open(yaml_file, "r") as file:
            completion_kwargs = yaml.safe_load(file)
        return cls(**completion_kwargs)

    def __call__(
        self,
        prompt: Union[str, List[Dict[str, str]]],
        completion_kwargs: Optional[Dict[str, Any]] = None,
    ) -> OpenAiChatResult:
        """Create a model response for the given prompt (chat conversation).

        Args:
            prompt: The prompt for the model.
            completion_kwargs: Keyword arguments for the OpenAI completion.

                - ``model`` can not be set via ``completion_kwargs``! Please set the ``model`` in the initializer.
                - ``messages`` can not be set via ``completion_kwargs``! Please set the ``prompt`` argument.

                Also see:

                    - ``openai.resources.chat.completions.Completions.create()``
                    - OpenAI API reference: `Create chat completion <https://platform.openai.com/docs/api-reference/chat/create>`_

        Returns:
            The result of the OpenAI completion.
        """
        if isinstance(prompt, list):
            for message in prompt:
                if "role" not in message or "content" not in message:
                    raise ValueError(
                        "If prompt is a list of messages, each message must have a 'role' and 'content' key!"
                    )
                if message["role"] not in ["system", "user", "assistant", "tool"]:
                    raise ValueError(
                        "If prompt is a list of messages, each message must have a 'role' key with one of the values "
                        "'system', 'user', 'assistant' or 'tool'!"
                    )

        if completion_kwargs is not None:
            # check keys of completion_kwargs
            if "model" in completion_kwargs:
                raise ValueError(
                    "'model' can not be set via 'completion_kwargs'! Please set the 'model' in the initializer."
                )
            if "messages" in completion_kwargs:
                raise ValueError(
                    "'messages' can not be set via 'completion_kwargs'! Please set the 'prompt' argument."
                )
        else:
            completion_kwargs = {}  # set default value
        completion_kwargs["model"] = self.model
        messages = [{"role": "user", "content": prompt}] if isinstance(prompt, str) else prompt
        chat_completion = self.client.chat.completions.create(
            messages=messages,  # type: ignore[arg-type]
            **completion_kwargs,
        )
        result = OpenAiChatResult.from_chat_completion(chat_completion, completion_kwargs=completion_kwargs)
        return result


@dataclass
class OpenAiAzureChat(OpenAiChat):
    """Tool to interact with Azure OpenAI chat models.

    This can also be constructed with :meth:`~OpenAiChat.from_yaml`.

    See Also:
        * OpenAI API reference: `Create chat completion <https://platform.openai.com/docs/api-reference/chat/create>`_
        * `Quickstart: Get started generating text using Azure OpenAI Service <https://learn.microsoft.com/en-us/azure/ai-services/openai/quickstart?tabs=command-line&pivots=programming-language-python>`_

    Args:
        api_key: The OpenAI API key.
        model: The OpenAI model name.
        api_version: The OpenAI API version.
            A common value for this is ``2023-05-15``.
        azure_endpoint: The Azure endpoint.
    """

    api_version: str
    azure_endpoint: str

    def __post_init__(self) -> None:
        """Do post init."""
        self.client = AzureOpenAI(
            api_key=self.api_key,
            api_version=self.api_version,
            azure_endpoint=self.azure_endpoint,
        )



# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers `SoMaJo <https://github.com/tsproisl/SoMaJo>`_ specific tools.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[somajo]``
"""


from abc import ABC
from dataclasses import dataclass, field
from typing import Container, Dict, Iterable, List, Literal, Optional, Set, Tuple, Union

from somajo import SoMaJo
from tqdm import tqdm


@dataclass
class SoMaJoBaseClass(ABC):
    """Base Class for SoMaJo tools.

    Args:
        language: The language. ``de_CMC`` for German or ``en_PTB`` for English.

    Note:
        This class is an abstract base class. It should not be used directly.
    """

    language: Literal["de_CMC", "en_PTB"]
    somajo: SoMaJo = field(init=False, repr=False)

    def __post_init__(self):
        """Do post init."""
        self.somajo = SoMaJo(self.language)


def detokenize(tokens) -> str:
    """Convert SoMaJo tokens to sentence (string).

    Args:
        tokens: The tokens to be de-tokenized.
    Returns:
        The de-tokenized sentence.

    See Also:
        `How do I split sentences but not words? <https://github.com/tsproisl/SoMaJo/issues/17>`_
    """
    result_list = []
    for token in tokens:
        if token.original_spelling is not None:
            result_list.append(token.original_spelling)
        else:
            result_list.append(token.text)

        if token.space_after:
            result_list.append(" ")
    result = "".join(result_list)
    result = result.strip()
    return result


def extract_token_class_set(sentences: Iterable, keep_token_classes: Optional[Container[str]] = None) -> Set[str]:
    """Extract token from sentences by token class.

    Args:
        sentences: The sentences from which to extract.
        keep_token_classes: The token classes to keep. If ``None`` all will be kept.
    Returns:
        The set of extracted token texts.
    """
    result = set()
    for sentence in sentences:
        for token in sentence:
            if keep_token_classes is None or token.token_class in keep_token_classes:
                result.add(token.text)
            # else ignore
    return result


@dataclass
class SoMaJoSentenceSplitter(SoMaJoBaseClass):
    """Use SoMaJo to split text into sentences.

    Args:
        language: The language. ``de_CMC`` for German or ``en_PTB`` for English.
        show_progress_bar: Show a progressbar during processing.
    """

    show_progress_bar: bool = False

    def __call__(self, text: str) -> List[str]:
        """Split the text into a list of sentences.

        Args:
            text: The text to be split.
        Returns:
            The list of sentence splits.
        """
        sentences = self.somajo.tokenize_text([text])

        result = []

        for sentence in tqdm(sentences, disable=not self.show_progress_bar):
            sentence_string = detokenize(sentence)
            result.append(sentence_string)

        return result


@dataclass
class JaccardSimilarity(SoMaJoBaseClass):
    """Calculate the `jaccard similarity <https://en.wikipedia.org/wiki/Jaccard_index>`_.

    Args:
        language: The language. ``de_CMC`` for German or ``en_PTB`` for English.
    """

    def get_token_set(self, text: str) -> Set[str]:
        """Get token set for text.

        Args:
            text: The text to be tokenized into a set.
        Returns:
            The set of tokens (words).
        """
        sentences = self.somajo.tokenize_text([text])
        token_set = extract_token_class_set(sentences)  # TODO: filter tokens
        token_set = {t.lower() for t in token_set}

        return token_set

    def __call__(self, text1: str, text2: str) -> float:
        """Calculate the jaccard similarity for two texts.

        Args:
            text1: Text one.
            text2: Text two.
        Returns:
            The jaccard similarity.
        """
        token_set1 = self.get_token_set(text1)
        token_set2 = self.get_token_set(text2)
        intersection = token_set1.intersection(token_set2)
        union = token_set1.union(token_set2)
        jaccard_similarity = float(len(intersection)) / len(union)
        return jaccard_similarity


@dataclass
class TokenExtractor(SoMaJoBaseClass):
    """Extract tokens from text.

    Args:
        language: The language. ``de_CMC`` for German or ``en_PTB`` for English.
    """

    def extract_url_set(self, text: Union[Iterable, str]) -> Set[str]:
        """Extract URLs from text.

        An example:

        .. testcode::

            from mltb2.somajo import TokenExtractor

            token_extractor = TokenExtractor("de_CMC")
            url_set = token_extractor.extract_url_set("Das ist ein Link: http://github.com")
            print(url_set)

        Example output:

        .. testoutput::

            {'http://github.com'}

        Args:
            text: the text
        Returns:
            Set of extracted links.
        """
        if isinstance(text, str):
            text = [text]
        sentences = self.somajo.tokenize_text(text)
        result = extract_token_class_set(sentences, keep_token_classes="URL")
        return result


@dataclass
class UrlSwapper:
    """Tool to swap (and reverse swap) links with a numbered replacement link.

    Args:
        token_extractor: The sentence token extractor to be used.
        url_pattern: The pattern to use for replacement. One ``{}`` marks the place where to put the number.
    """

    token_extractor: TokenExtractor
    url_pattern: str = "https://link-{}.com"
    _url_map: Dict[str, str] = field(init=False, repr=False)  # map from real url to swapped url

    def __post_init__(self):
        """Do post init."""
        self._url_map = {}

    def swap_urls(self, text: str) -> str:
        """Swap the urls of the text."""
        url_set = self.token_extractor.extract_url_set(text)
        for url in url_set:
            if url not in self._url_map:  # if url is unknown: add it
                self._url_map[url] = self.url_pattern.format(len(self._url_map) + 1)
            text = text.replace(url, self._url_map[url])  # replace
        return text

    def reverse_swap_urls(self, text: str) -> Tuple[str, Set[str]]:
        """Revert the url swap.

        Returns:
            The reverted text and a ``set`` of URLs that were unknown by the ``URLSwapper``.
        """
        reverse_url_map = {v: k for k, v in self._url_map.items()}  # map from swapped url to real url
        url_set = self.token_extractor.extract_url_set(text)
        no_reverse_swap_urls = set()
        for url in url_set:
            if url in reverse_url_map:
                text = text.replace(url, reverse_url_map[url])  # replace
            else:
                no_reverse_swap_urls.add(url)
        return text, no_reverse_swap_urls



# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""This module offers text specific tools.

It offers the following functionality:

- detect or clean invisible characters
- detect or replace special whitespaces
- remove duplicate whitespaces
- calculate the distance between two texts to find anomalies
"""

import re
from collections import Counter, defaultdict
from dataclasses import dataclass, field
from typing import Dict, Final, Iterable, Optional, Pattern, Set, Tuple, Union

from scipy.spatial.distance import cityblock
from tqdm import tqdm

INVISIBLE_CHARACTERS: Final[Tuple[str, ...]] = (
    "\u200b",  # Zero Width Space (ZWSP) https://www.compart.com/en/unicode/U+200b
    "\u00ad",  # Soft Hyphen (SHY) https://www.compart.com/en/unicode/U+00ad
    # TODO: what about:
    # https://www.compart.com/en/unicode/U+2028
    # https://www.compart.com/en/unicode/U+2029
)

INVISIBLE_CHARACTERS_TRANS: Final[Dict[int, None]] = str.maketrans({char: None for char in INVISIBLE_CHARACTERS})

SPECIAL_WHITESPACES: Final[Tuple[str, ...]] = (
    # unicode block "General Punctuation": https://www.compart.com/en/unicode/block/U+2000
    "\u2000",  # En Quad
    "\u2001",  # Em Quad
    "\u2002",  # En Space
    "\u2003",  # Em Space
    "\u2004",  # Three-Per-Em Space
    "\u2005",  # Four-Per-Em Space
    "\u2006",  # Six-Per-Em Space
    "\u2007",  # Figure Space https://www.compart.com/en/unicode/U+2007
    "\u2008",  # Punctuation Space
    "\u2009",  # Thin Space https://www.compart.com/en/unicode/U+2009
    "\u200a",  # Hair Space https://www.compart.com/en/unicode/U+200A
    "\u202f",  # Narrow No-Break Space (NNBSP) https://www.compart.com/en/unicode/U+202f
    # other unicode blocks
    "\u00a0",  # No-Break Space (NBSP) https://www.compart.com/en/unicode/U+00a0
)

SPECIAL_WHITESPACES_TRANS: Final[Dict[int, str]] = str.maketrans({char: " " for char in SPECIAL_WHITESPACES})

INVISIBLE_CHARACTERS_AND_SPECIAL_WHITESPACES_TRANS = {**SPECIAL_WHITESPACES_TRANS, **INVISIBLE_CHARACTERS_TRANS}

MULTI_SPACE_PATTERN: Pattern = re.compile(r" {2,}")

XML_TAG_PATTERN: Pattern = re.compile(r"<\/?[\w:]+( \/|\/|)>")


def has_xml_tag(text: str) -> bool:
    """Check if text contains XML tags (one or multiple).

    These are some XML tags we detect:

    - ``<xml_tag>``
    - ``<xml:tag>``
    - ``</xml_tag>``
    - ``<xml_tag/>``
    - ``<xml_tag />``

    While we do not detect ``a < b but x > y``.

    Args:
        text: The text to check.
    Returns:
        ``True`` if the text contains XML tags, ``False`` otherwise.
    """
    return re.search(XML_TAG_PATTERN, text) is not None


def remove_invisible_characters(text: str) -> str:
    """Remove invisible characters from text.

    The invisible characters are defined in the constant ``INVISIBLE_CHARACTERS``.

    Args:
        text: The text from which the invisible characters are to be removed.

    Returns:
        The cleaned text.
    """
    return text.translate(INVISIBLE_CHARACTERS_TRANS)


def has_invisible_characters(text: str) -> bool:
    """Check if text contains invisible characters.

    The invisible characters are defined in the constant ``INVISIBLE_CHARACTERS``.

    Args:
        text: The text to check.

    Returns:
        ``True`` if the text contains invisible characters, ``False`` otherwise.
    """
    return any(char in text for char in INVISIBLE_CHARACTERS)


def replace_special_whitespaces(text: str) -> str:
    """Replace special whitespaces with normal whitespaces.

    The special whitespaces are defined in the constant ``SPECIAL_WHITESPACES``.

    Args:
        text: The text from which the special whitespaces are to be replaced.

    Returns:
        The cleaned text.
    """
    return text.translate(SPECIAL_WHITESPACES_TRANS)


def has_special_whitespaces(text: str) -> bool:
    """Check if text contains special whitespaces.

    The special whitespaces are defined in the constant ``SPECIAL_WHITESPACES``.

    Args:
        text: The text to check.

    Returns:
        ``True`` if the text contains special whitespaces, ``False`` otherwise.
    """
    return any(char in text for char in SPECIAL_WHITESPACES)


def replace_multiple_whitespaces(text: str) -> str:
    """Replace multiple whitespaces with single whitespace.

    Args:
        text: The text from which the multiple whitespaces are to be replaced.

    Returns:
        The cleaned text.
    """
    return MULTI_SPACE_PATTERN.sub(" ", text)


def clean_all_invisible_chars_and_whitespaces(text: str) -> str:
    """Clean text form invisible characters and whitespaces.

    - Remove invisible characters from text.
    - Replace special whitespaces with normal whitespaces.
    - Replace multiple whitespaces with single whitespace.
    - Remove leading and trailing whitespaces.

    The invisible characters are defined in the constant ``INVISIBLE_CHARACTERS``.
    The special whitespaces are defined in the constant ``SPECIAL_WHITESPACES``.

    Args:
        text: The text to clean.

    Rteturns:
        The cleaned text.
    """
    text = text.translate(INVISIBLE_CHARACTERS_AND_SPECIAL_WHITESPACES_TRANS)
    text = replace_multiple_whitespaces(text)
    text = text.strip()
    return text


def _normalize_counter_to_defaultdict(counter: Counter, max_dimensions: int) -> defaultdict:
    """Normalize a counter to to ``max_dimensions``.

    The number of dimensions is limited to ``max_dimensions``
    of the most commen characters.
    The counter values are normalized by deviding them by the total count.

    Args:
        counter: The counter to normalize.
        max_dimensions: The maximum number of dimensions to use for the normalization.
            Must be greater than 0.
    Returns:
        The normalized counter with a maximum of ``max_dimensions`` dimensions.
    """
    total_count = sum(counter.values())
    normalized_counter = defaultdict(float)
    for char, count in counter.most_common(max_dimensions):
        normalized_counter[char] = count / total_count
    return normalized_counter


@dataclass
class TextDistance:
    """Calculate the distance between two texts.

    This class can be used to find texts with anomalies.
    For example with HTML markup or other unusual characters.

    One text (or multiple texts) must first be fitted with :func:`~TextDistance.fit`.
    After that the distance to other given texts can be calculated with :func:`~TextDistance.distance`.
    After the distance was calculated the first time, the class can
    not be fitted again.

    Args:
        show_progress_bar: Show a progressbar during processing.
        max_dimensions: The maximum number of dimensions to use for the distance calculation.
            Must be greater than 0.
    Raises:
        ValueError: If ``max_dimensions`` is not greater than 0.
    """

    show_progress_bar: bool = False
    max_dimensions: int = 100

    # counter for the text we fit
    _char_counter: Optional[Counter] = field(default_factory=Counter, init=False)

    # normalized counter for the text we fit - see _normalize_char_counter
    _normalized_char_counts: Optional[defaultdict] = field(default=None, init=False)

    # set of all counted characters - see _normalize_char_counter
    _counted_char_set: Optional[Set[str]] = field(default=None, init=False)

    # flag if fit was called
    _fit_called: bool = field(default=False, init=False)

    # flag if distance was called
    _distance_called: bool = field(default=False, init=False)

    def __post_init__(self) -> None:
        """Do post init."""
        if not self.max_dimensions > 0:
            raise ValueError("'max_dimensions' must be > 0!")

    def fit(self, text: Union[str, Iterable[str]]) -> None:
        """Fit the text.

        This method must be called at least once before :func:`~TextDistance.distance`.

        Args:
            text: The text to fit.
        Raises:
            ValueError: If :func:`~TextDistance.fit` is called after
                :func:`~TextDistance.distance`.
        """
        if self._distance_called:
            raise ValueError("fit must not be called after distance calculation!")

        if isinstance(text, str):
            self._char_counter.update(text)  # type: ignore[union-attr]
        else:
            for t in tqdm(text, disable=not self.show_progress_bar):
                self._char_counter.update(t)  # type: ignore[union-attr]

        self._fit_called = True

    def _normalize_char_counter(self) -> None:
        """Normalize the char counter to a defaultdict.

        This supports lazy postprocessing of the char counter.
        """
        if not self._distance_called:
            self._normalized_char_counts = _normalize_counter_to_defaultdict(self._char_counter, self.max_dimensions)  # type: ignore[arg-type]
            self._char_counter = None
            self._counted_char_set = set(self._normalized_char_counts)
            self._distance_called = True

    def distance(self, text) -> float:
        """Calculate the distance between the fitted text and the given text.

        This implementation uses the Manhattan distance (:func:`scipy.spatial.distance.cityblock`).
        The distance is only calculated for ``max_dimensions`` most commen characters.

        Args:
            text: The text to calculate the Manhattan distance to.
                The higher this value is, the more the text differs from the fitted text.
        """
        if not self._fit_called:
            raise ValueError("fit must not be called before distance!")
        self._normalize_char_counter()
        all_vector = []
        text_vector = []
        text_count = Counter(text)
        text_count_defaultdict = _normalize_counter_to_defaultdict(text_count, self.max_dimensions)
        for c in self._counted_char_set.union(text_count_defaultdict):  # type: ignore[union-attr]
            all_vector.append(
                self._normalized_char_counts[c]  # type: ignore[index]
            )  # if c is not in defaultdict, it will return 0
            text_vector.append(text_count_defaultdict[c])  # if c is not in defaultdict, it will return 0
        return cityblock(all_vector, text_vector)



# Copyright (c) 2023 Philip May, Deutsche Telekom AG
# Copyright (c) 2023-2024 Philip May
# This software is distributed under the terms of the MIT license
# which is available at https://opensource.org/licenses/MIT

"""Markdown specific module.

Hint:
    Use pip to install the necessary dependencies for this module:
    ``pip install mltb2[md]``
"""

import re
from dataclasses import dataclass
from typing import Final, List

from tqdm import tqdm

from mltb2.transformers import TransformersTokenCounter

_HEADLINE_REGEX: Final = re.compile(r"^#+ .*", flags=re.MULTILINE)


def _chunk_md_by_headline(md_text: str) -> List[str]:
    """Chunk Markdown by headlines.

    Args:
        md_text: The Markdown text to be chunked.

    Returns:
        The list of Markdown chunks.
    """
    positions: List[int] = [m.start() for m in re.finditer(_HEADLINE_REGEX, md_text)]

    # extend positions
    if 0 not in positions:
        positions = [0] + positions  # noqa: RUF005
    positions.append(len(md_text))

    result = [md_text[x:y].strip() for x, y in zip(positions, positions[1:])]
    return result


def chunk_md(md_text: str) -> List[str]:
    """Chunk Markdown by headlines and merge isolated headlines.

    Merges isolated headlines with their corresponding subsequent paragraphs.
    Headings isolated at the end of ``md_text`` (headings without content) are removed in this process.

    Args:
        md_text: The Markdown text to be chunked.

    Returns:
        The list of Markdown chunks.
    """
    md_chunks = _chunk_md_by_headline(md_text)

    merged_chunks = []
    temp_merged_chunk = []
    for chunk in md_chunks:
        temp_merged_chunk.append(chunk)
        if "\n" in chunk:  # content chunk found
            joined_content = "\n\n".join(temp_merged_chunk)
            merged_chunks.append(joined_content)
            temp_merged_chunk = []

    # if len(temp_content) > 0 this is only headlines and we skip them
    return merged_chunks


@dataclass
class MdTextSplitter:
    """Split Markdown text into sections with a specified maximum token number.

    Does not divide headings with their corresponding paragraphs.

    Args:
        max_token: Maximum number of tokens per text section.
            Can only be exceeded if a single Markdown chunk is already larger.
        transformers_token_counter: The token counter to be used.
        show_progress_bar: Show a progressbar during processing.
    """

    max_token: int
    transformers_token_counter: TransformersTokenCounter
    show_progress_bar: bool = False

    def __call__(self, md_text: str) -> List[str]:
        """Split the Markdown text into sections.

        Args:
            md_text: The Markdown text to be split.
        Returns:
            The list of Markdown section splits.
        """
        md_chunks = chunk_md(md_text)
        counts = self.transformers_token_counter(md_chunks)

        assert len(md_chunks) == len(counts)  # type: ignore[arg-type]

        result_merges: List[str] = []
        temp_merges: List[str] = []
        current_count: int = 0

        for md_chunk, count in zip(
            tqdm(md_chunks, disable=not self.show_progress_bar), counts  # type: ignore[arg-type]
        ):
            if current_count + count > self.max_token and len(temp_merges) > 0:
                joined_content = "\n\n".join(temp_merges)
                result_merges.append(joined_content)
                temp_merges = []
                current_count = 0

            current_count += count
            temp_merges.append(md_chunk)

        # add the rest
        if len(temp_merges) > 0:
            joined_content = "\n\n".join(temp_merges)
            result_merges.append(joined_content)

        return result_merges
